#========================================================================================
#  rompattobytes.pl
#  Copyright (C) 2003-2020 Makoto Kamada
#
#  This file is part of the XEiJ (X68000 Emulator in Java).
#  You can use, modify and redistribute the XEiJ if the conditions are met.
#  Read the XEiJ License for more details.
#  https://stdkmd.net/xeij/
#========================================================================================

use strict;  #厳密な文法に従う
use warnings;  #警告を表示する
use utf8;  #UTF-8で記述する

my $TEXT_START = 0x00fe4000;  #ROMデバッガの後ろが48KB空いている
my $TEXT_END = 0x00fef400;  #$00FEF400～$00FEFFFFに6x14 ANKフォントを入れる

{
  my ($name, $file) = @ARGV;
  #読み込む
  open IN, '<', $file or die;
  binmode IN;
  my ($head_buffer, $text_buffer, $data_buffer);
  read IN, $head_buffer, 64;
  my $text_size = vec $head_buffer, 3, 32;
  my $data_size = vec $head_buffer, 4, 32;
  read IN, $text_buffer, $text_size;
  read IN, $data_buffer, $data_size;
  close IN;
  #(挿入コードの長さ)<=(挿入終了アドレス-挿入開始アドレス)を確認する
  if ($data_size != 0) {
    my $pointer = 0;
    while (1) {
      my $start = unpack 'N', substr $data_buffer, $pointer, 4;
      $start == 0 and last;
      my ($end, $original, $size) = unpack 'N3', substr $data_buffer, $pointer + 4, 12;
      printf "  //  data: \$%08X-\$%08X %d/%d %s\n",
      $start, $end - 1, $size, $end - $start,
      $size <= $end - $start ? 'OK' : 'ERROR';
      $pointer += 16 + $size;
    }
  }
  #
  my $text_start = vec $head_buffer, 1, 32;
  $text_start == $TEXT_START or die "ERROR: base address mismatch. modify Makefile.x68k\nTEXT_START = " . sprintf ('%x', $TEXT_START) . "\n";
  my $text_end = $TEXT_END;
  if ($text_size != 0) {
    printf "  //  text: \$%08X-\$%08X %d/%d %s\n",
    $text_start, $text_end - 1, $text_size, $text_end - $text_start,
    $text_size <= $text_end - $text_start ? 'OK' : 'ERROR';
    printf "  public static final int %s_BASE = 0x%08x;\n", $name, $text_start;
  }
  foreach my $pair (['TEXT', $text_buffer], ['DATA', $data_buffer]) {
    length $pair->[1] or next;
    my $str = join '', map {
      $_ == 8 ? '\\b' :
      $_ == 9 ? '\\t' :
      $_ == 10 ? '\\n' :
      $_ == 12 ? '\\f' :
      $_ == 13 ? '\\r' :
      $_ == 34 || $_ == 39 || $_ == 92 ? sprintf '\\%c', $_ :
      32 <= $_ && $_ <= 126 ? chr $_ :
      sprintf '\\%03o', $_;
    } unpack 'C*', $pair->[1];
    $str =~ s/(?<!\\)((?:\\\\)*\\)00([0-7])(?![0-7])/$1$2/g;
    $str =~ s/(?<!\\)((?:\\\\)*\\)0([0-7]{2})(?![0-7])/$1$2/g;
    printf "  public static final byte[] %s_%s = \"%s\".getBytes (XEiJ.ISO_8859_1);\n", $name, $pair->[0], $str;
  }
}
